from functools import partial

from typing import (
    Any,
)

from torch.utils.data import DataLoader

import torch.nn as nn

from torch import Tensor

from .strategy import DataParalelism
from .environment import SingleNodeEnvironment
from .dataloader import _DistributonDataLoader

class Distributon:
    """
    We have three main components: the launcher, the strategy and the environment.
    """
    def __init__(self, devices):
        self._num_nodes = 1
        self.num_processes = len(devices)      
        self._strategy = DataParalelism(devices, SingleNodeEnvironment())


    def launch(self, *args: Any, **kwargs: Any):
        """Launch and initialize the processes needed for the distributed training. """
        to_run = partial(self._strategy.setup_environment)
        self._strategy.launcher.launch(to_run, args, kwargs)


    def setup_optimizer(self, optimizer):
        """ Returns wrapper over optimizer """

        optimizer = self._strategy.setup_optimizer(optimizer)
        # TODO: we need to add a wrapper over the optimizer
        # to delegate optimizer step calls to the strategy

        return optimizer

    def setup_model(self, model: nn.Module) -> Any:
        """ Returns wrapper over model """
        # Move the model to the device
        self._strategy.model_to_device(model=model)
        model = self._strategy.setup_module(model)

        return model


    def setup_dataloaders(self, dataloader: DataLoader):
        """ Returns wrapper over dataloadewr """
        distributon_dataloader = _DistributonDataLoader(dataloader=dataloader, device=self._strategy.root_device)
        return distributon_dataloader


    def backward(self, tensor: Tensor, *args: Any, **kwargs: Any) -> None:
        r"""Forwards backward-calls to the precision plugin."""
        # TODO: Use mixed precission for this
        tensor.backward(*args, **kwargs)
